
import casadi.*
%% discretization
N = 20;
T = 1; % time horizon length
dT = T/N;

x0 = [0; pi; 0; 0];
U_max = 80;

%% model dynamics
model = pendulum_on_cart_model;
nx = model.nx;
nu = model.nu;
ny = size(model.cost_expr_y, 1);      % used in simulink example
ny_e = size(model.cost_expr_y_e, 1);

Q = model.W(1:nx, 1:nx);
R = model.W(end-nu+1:end, end-nu+1:end);

Q_e = model.W_e;

%% RK4 
x = model.sym_x;
u = model.sym_u;
f_continuous = Function('f_continuous', {x, u}, {model.expr_f_expl});

k1 = f_continuous(x,       u);
k2 = f_continuous(x+dT/2*k1,u);
k3 = f_continuous(x+dT/2*k2,u);
k4 = f_continuous(x+dT*k3,  u);
xf = x + dT/6 * (k1 + 2*k2 + 2*k3 + k4);

f_discrete = Function('f_discrete', {model.sym_x, model.sym_u}, {xf});

%% setup ocp

x_traj = SX.sym('x_traj', nx*(N+1), 1);
u_traj = SX.sym('u_traj', nu*N, 1);

xs = vertsplit(x_traj, 0:nx:nx*(N+1));
us = vertsplit(u_traj, 0:nu:nu*N);

objective = 0;
continuity_constraints = {};

for n=1:N
    u_current = us{n};
    x_current = xs{n};
    x_next = xs{n+1};
    continuity_constraints = [continuity_constraints, ...
                  {x_next - f_discrete(x_current, u_current)}];

    objective = objective + x_current'*Q*x_current + u_current'*R*u_current;
end

x_final = xs{N+1};

objective = objective + x_final'*Q_e*x_final;
objective = 0.5 * objective;

continuity_constraints = vertcat(continuity_constraints{:});
x0_bar = SX.sym('x0_bar', nx, 1);

initial_constraint = xs{1} - x0_bar;

%% solver
nlp = struct('x', vertcat(x_traj, u_traj), 'p', x0_bar, 'f', objective, ...
             'g', vertcat(initial_constraint, continuity_constraints, u_traj));
         
opts = struct('jit', true, 'jit_options', struct('flags', '-O3'), ...
              'ipopt', struct('print_level', 0, 'tol', 1e-6), ...
              'print_time', false, 'record_time', true);

solver = nlpsol('solver', 'ipopt', nlp, opts);

%% solve
initial_guess = zeros(size(nlp.x));

lbg = zeros(size(nlp.g));
ubg = zeros(size(nlp.g));
lbg(end-nu*N+1:end) = -U_max;
ubg(end-nu*N+1:end) = +U_max;

n_executions = 10;

times_casadi = zeros(n_executions, 1);

for i=1:n_executions
    res = solver('x0', initial_guess, 'p', x0, 'lbg', lbg, 'ubg', ubg);
    times_casadi(i) =  solver.stats.t_wall_total;
end
%%
disp(['Average time casadi: ' num2str(sum(times_casadi)/n_executions*1000), 'ms']);


opt = full(res.x);
ts = linspace(0, T, N+1);
u_opt =  opt(end-nu*N+1:end);
figure
stairs(ts(1:end-1), u_opt)
ylabel('F [N]')
xlabel('t [s]')
grid on
